"""Utilities for concurrency of Python functions with multithreading or multiprocessing."""
import concurrent.futures
import functools
import typing

from ori.errors import OriValidationError


def run_in_background(
    func: typing.Callable,
    *,
    executor_class: type,
    max_workers: typing.Optional[int] = None,
):
    """
    Runs the function in a background thread or process, depending on the
    executor that you provided.

    Most users will want to use :meth:`ori.concurrency.run_in_background_thread` or
    :meth:`ori.concurrency.run_in_background_process` instead.

    Args:
        func: The function to run in the background.
        executor_class: A `concurrent.futures.Executor` subclass that the
            function will be submitted to. Two popular choices are
            :class:`concurrent.futures.ThreadPoolExecutor` or
            :class:`concurrent.futures.ProcessPoolExecutor`, but you can pass
            your own subclass if if it follows the same API.
        max_workers: Specify this to set the maximum number of parallel
            threads or processes to run at once. If this is set to
            None, then the executor's default maximum number of workers
            are used.

    Returns:
        This function returns a wrapper of the `func` function that you passed.
        The wrapper takes the exact same arguments, but instead returns
        a :class:`concurrent.futures.Future` object. To wait for the return
        value of `func`, call `.result()` on the future that this function
        returns.

    Raises:
        OriValidationError: When you provide a value for `func` that is not
            a callable function.

    """
    if not callable(func):
        raise OriValidationError(
            "You need to provide a callable function as the first " "argument."
        )
    executor = executor_class(max_workers=max_workers)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return executor.submit(func, *args, **kwargs)

    wrapper.executor = executor
    return wrapper


def run_in_background_thread(
    func: typing.Callable, *, max_workers: typing.Optional[int] = None,
):
    """
    Runs the function in a background thread.

    Args:
        func: The function to run in the background.
        max_workers: Specify this to set the maximum number of parallel
            threads or processes to run at once. If this is set to
            None, then the executor's default maximum number of workers
            are used.

    Returns:
        This function returns a wrapper of the `func` function that you passed.
        The wrapper takes the exact same arguments, but instead returns
        a :class:`concurrent.futures.Future` object. To wait for the return
        value of `func`, call `.result()` on the future that this function
        returns.

    Raises:
        OriValidationError: When you provide a value for `func` that is not
            a callable function.

    """
    return run_in_background(
        func=func,
        executor_class=concurrent.futures.ThreadPoolExecutor,
        max_workers=max_workers,
    )


def run_in_background_process(
    func: typing.Callable, *, max_workers: typing.Optional[int] = None,
):
    """
    Runs the function in a background process.

    Args:
        func: The function to run in the background.
        max_workers: Specify this to set the maximum number of parallel
            threads or processes to run at once. If this is set to
            None, then the executor's default maximum number of workers
            are used.

    Returns:
        This function returns a wrapper of the `func` function that you passed.
        The wrapper takes the exact same arguments, but instead returns
        a :class:`concurrent.futures.Future` object. To wait for the return
        value of `func`, call `.result()` on the future that this function
        returns.

    Raises:
        OriValidationError: When you provide a value for `func` that is not
            a callable function.
    """
    return run_in_background(
        func=func,
        executor_class=concurrent.futures.ProcessPoolExecutor,
        max_workers=max_workers,
    )



"""All exceptions for the Ori module."""


class OriError(Exception):
    """The base class efor all Ori errors."""


class OriValidationError(OriError):
    """Raise when Ori receives bad input from the user."""



"""
Module for the poolchain.

The "poolchain" is a high-level abstraction over the Python
:mod:`concurrent.futures` concurrency  module. A poolchain used
to run a single iterable (e.g. a list) through a chain of functions 
and collect an iterable of results at the end.

For example, here is one way to use a poolchain.

.. highlight:: python
.. code-block:: python

    from ori.poolchain import PoolChain

    results = (
        PoolChain()
        .add_threadpool(lambda num: num * 10)
        .add_threadpool(lambda num: str(num))
        .add_threadpool(lambda s: s + " is a number")
    ).execute_lazy([1, 2, 3, 4, 5])

    for row in results:
        print(row)
    # `results` is an iterator, so we go through it with a for loop.
    # we could have run `.execute_eager()` to automatically
    # turn the iterator into a list.


The general workflow is to create a :mod:`PoolChain` object and
then make a series of chained calls that add functions to execute
over every item in the iterable in sequence.

After the chain of functions has been specified, call one of the
`.execute_*()` functions like :func:`PoolChain.execute_lazy`
to get an iterator of the final return value or
:func:`PoolChain.execute_eager` to get an entire list upfront.


"""

import concurrent.futures
import typing

from ori.errors import OriValidationError


class _ChainElement(typing.NamedTuple):
    """A single element in our PoolChain."""

    function: typing.Callable
    executor_class: type
    max_workers: typing.Optional[int]
    timeout: typing.Optional[int]
    chunksize: int


class PoolChain:
    """Creates a poolchain."""

    def __init__(self):
        """Create a new poolchain."""
        self.chain: typing.Iterable[_ChainElement] = []

    def add(
        self,
        function: typing.Callable,
        *,
        executor_class: type,
        max_workers: typing.Optional[int] = None,
        timeout: typing.Optional[int] = None,
        chunksize: int = 1,
    ):
        """
        Add a new element to the poolchain.

        Note that if your `executor_class` is a
        :class:`concurrent.futures.ProcessPoolExecutor`, you will not be
        able to pass `lambda` functions as `func`. You will need to use
        named functions defined with the `def` keyword.

        Args:
            function: This is the function that you want to run at the
                end of the current chain.
            executor_class: the :class:`concurrent.futures.Executor` subclass
                type to use to execute this part of the poolchain.
            max_workers: The maximum number of workers that the
                :class:`concurrent.futures.ProcessPoolExecutor` can start.
                If this is `None`, then then the executor will create
                the same number of workers as you have processors on
                your machine. This is usually a sensible default.
            timeout: The amount of time (in seconds) to wait for a
                :class:`concurrent.futures.Future` before throwing an error.
            chunksize: Sometimes the executor runs faster when providing chunks
                of the iterable to workers rather than just one iterable
                element at a time (the default). Try setting chunksize to
                a value greater than one to more efficiently use
                interprocess communication.

        Returns:
            The PoolChain object itself, allowing you to chain subsequent
            `add_*()` calls and at the end an `execute_*()` call.

        Raises:
            OriValidationError: Thrown if you don't send a real function, or
                a real Executor, a non-integer for one of the integer fields,
                etc.
        """
        if not isinstance(executor_class, type):
            raise OriValidationError(
                "For the `executor_class` argument, you should pass a Python "
                "class. Did you pass a Python instance instead?"
            )
        if not getattr(executor_class, "map", False):
            raise OriValidationError(
                "Your `executor_class` needs to implement the `.map` "
                "method. It looks like it does not."
            )
        if (
            function.__name__ == "<lambda>"
            and executor_class == concurrent.futures.ProcessPoolExecutor
        ):
            raise OriValidationError(
                "You cannot use lambda functions with ProcessPoolExecutor. "
                "Define a named function with the `def` keyword instead."
            )
        if max_workers is not None:
            if not isinstance(max_workers, int):
                raise OriValidationError(
                    "The `max_workers` argument should be an int or None."
                )
            if max_workers < 1:
                raise OriValidationError(
                    "You need to define one or more worker. "
                    f"You defined {max_workers} workers."
                )
        if timeout is not None:
            if not isinstance(timeout, int):
                raise OriValidationError("Timeout should be None or an integer.")
            if timeout < 1:
                raise OriValidationError(
                    "Timeout should be None or an integer greater than 0."
                )
        self.chain.append(
            _ChainElement(
                function=function,
                executor_class=executor_class,
                max_workers=max_workers,
                timeout=timeout,
                chunksize=chunksize,
            )
        )
        return self

    def add_threadpool(
        self,
        function: typing.Callable,
        *,
        max_workers: typing.Optional[int] = None,
        timeout: typing.Optional[int] = None,
    ):
        """Add a function to run with threadpool to the poolchain.

        Args:
            function: This is the function that you want to run at the
                end of the current chain.
            max_workers: The maximum number of workers that the
                :class:`concurrent.futures.ProcessPoolExecutor` can start.
                If this is `None`, then then the executor will create
                the same number of workers as you have processors on
                your machine. This is usually a sensible default.
            timeout: The amount of time (in seconds) to wait for a
                :class:`concurrent.futures.Future` before throwing an error.

        Returns:
            The PoolChain object itself, allowing you to chain subsequent
            `add_*()` calls and at the end an `execute_*()` call.

        Raises:
            OriValidationError: Thrown if you don't send a real function, or
                a real Executor, a non-integer for one of the integer fields,
                etc.
        """
        return self.add(
            function=function,
            executor_class=concurrent.futures.ThreadPoolExecutor,
            max_workers=max_workers,
            timeout=timeout,
        )

    def add_processpool(
        self,
        function: typing.Callable,
        *,
        max_workers: typing.Optional[int] = None,
        timeout: typing.Optional[int] = None,
        chunksize: int = 1,
    ):
        """
        Add a function to run with a "process pool" to the poolchain.
        
        You cannot pass a `lambda` function to a process pool. You need
        to make up your poolchain entirely with named functions defined with
        the `def` keyword.

        Args:
            function: This is the function that you want to run at the
                end of the current chain.
            max_workers: The maximum number of workers that the
                :class:`concurrent.futures.ProcessPoolExecutor` can start.
                If this is `None`, then then the executor will create
                the same number of workers as you have processors on
                your machine. This is usually a sensible default.
            timeout: The amount of time (in seconds) to wait for a
                :class:`concurrent.futures.Future` before throwing an error.
            chunksize: Sometimes the executor runs faster when providing chunks
                of the iterable to workers rather than just one iterable
                element at a time (the default). Try setting chunksize to
                a value greater than one to more efficiently use
                interprocess communication.

        Returns:
            The PoolChain object itself, allowing you to chain subsequent
            `add_*()` calls and at the end an `execute_*()` call.

        Raises:
            OriValidationError: Thrown if you don't send a real function, or
                a real Executor, a non-integer for one of the integer fields,
                etc.
        """
        return self.add(
            function=function,
            executor_class=concurrent.futures.ProcessPoolExecutor,
            max_workers=max_workers,
            timeout=timeout,
            chunksize=chunksize,
        )

    def execute_lazy(self, iterable: typing.Iterable[typing.Any]):
        """Process the given iterable through the poolchain.

        When you call this function, the poolchain begins executing
        immediately

        Returns:
            This function returns an iterator that lazily executes
            the poolchain.
        """
        if len(self.chain) < 1:
            raise OriValidationError("Try adding some workers first.")

        current_iterable = iterable
        executors = []
        for element in self.chain:
            current_executor = element.executor_class(max_workers=element.max_workers,)
            executors.append(current_executor)
            current_iterable = current_executor.map(
                element.function,
                current_iterable,
                timeout=element.timeout,
                chunksize=element.chunksize,
            )
        for item in current_iterable:
            yield item
        for executor in executors:
            executor.shutdown(wait=True)

    def execute_eager(self, iterable: typing.Iterable[typing.Any]):
        """
        Run the given iterable through the poolchain, eagerly returning a list.

        This method returns the results upfront in a list, which lets you
        look at the entire output without having to iterate through it yourself.

        Args:
            iterable: This is an iterable where every element in the
                iterable is processed through the chain.

        Returns:
            This function returns a complete Python list containing the return
            value of every element in the list.
        """
        return list(self.execute_lazy(iterable))

    def execute_lazy_single_threaded(self, iterable: typing.Iterable[typing.Any]):
        """
        Execute the chain in the current thread, returning a generator.

        This function is typically used when you want to test your chain without
        creating threads or processes. When running in one thread,
        :mod:`PoolChain` would never be faster than just running the functions
        yourself in sequence.

        This function returns a generator that you have to iterate through
        yourself for results.

        Args:
            iterable: This is an iterable where every element in the
                iterable is processed through the chain.

        Returns:
            This function returns an iterator that lazily executes
            the poolchain in the current thread. Every time the next
            output is fetched, the corresponding input is sent through
            every function in the chain.
        
        """
        for item in iterable:
            current_val = item
            for element in self.chain:
                current_val = element.function(current_val)
            yield current_val

    def execute_eager_single_threaded(self, iterable: typing.Iterable[typing.Any]):
        """
        Execute the chain in the current thread, returning a Python list.

        This method is typically used when you want to test your chain without
        creating threads or processes. When running in one thread,
        :mod:`PoolChain` would never be faster than just running the functions
        yourself in sequence.

        This method also returns the results upfront in a list, which lets you
        look at the entire output without having to iterate through it yourself.

        Args:
            iterable: This is an iterable where every element in the
                iterable is processed through the chain.

        Returns:
            This function returns a complete Python list containing the return
            value of every element in the list.
            
        """
        return list(self.execute_lazy_single_threaded(iterable))



"""Utility function for running external commands as subprocesses."""
import multiprocessing
import selectors
import signal
import subprocess
import sys
import typing


def _null_function(*args, **kwargs):  # pylint: disable=unused-argument
    """A function that does nothing.

    This function is used to replace printing or logging functions
    when they are not specified.
    """


def run_process_in_background(
    *,
    command: typing.Iterable[str],
    stdout_function: typing.Optional[typing.Callable] = None,
    stderr_function: typing.Optional[typing.Callable] = None,
    exception_function: typing.Optional[typing.Callable] = None,
    exit_code_on_killed: int = 0,
):
    """
    Runs `command`, writing stdout and stderr to the logger in `logger`.

    This function is based on `this answer on StackOverflow <https://stackoverflow.com/questions/31833897/python-read-from-subprocess-stdout-and-stderr-separately-while-preserving-order/56918582#56918582>`_.

    Args:
        command: This is an iterable--like a list--of strings. This iterable
            describes a command line program to run. For example, the
            command `ls -l /home` would be broken up into
            `["ls", "-l", "/home"]`. You can use the function
            :func:`shlex.split` to turn any string command into an iterable.
        stdout_function: This is a function that your subprocess will call for
            every line of standard output that your program emits.
            You can choose functions like :func:`print` or, say,
            :func:`logging.debug`.
        stderr_function: This is a function that your subprocess will call
            for every line of standard error that your program emits.
            The same rules apply as for `stdout_function`.
        exception_function: This is a function to call if we catch an exception
            while logging the program.
        exit_code_on_killed: This is the exit code we return for the subprocess
            when we catch an exception while logging the program.

    Returns:
        This returns an already-started :class:`multiprocessing.Process` instance,
        which you can use to monitor or kill the process as time goes on.
    """

    if not stdout_function:
        stdout_function = _null_function
    if not stderr_function:
        stderr_function = _null_function
    if not exception_function:
        exception_function = _null_function

    def target():
        """Manage the actual command we're running."""
        command_process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True,
        )
        try:
            # If the control process receives a signal to stop,
            # then make sure to stop the command process.
            def kill_handler(signum, frame):  # pylint: disable=unused-argument
                command_process.terminate()
                sys.exit(exit_code_on_killed)

            signal.signal(signal.SIGHUP, kill_handler)
            signal.signal(signal.SIGTERM, kill_handler)
            signal.signal(signal.SIGINT, kill_handler)

            # Listen on stdout and stderr in parallel, preserving order,
            # writing stdout to debug logs and stderr to info logs.
            sel = selectors.DefaultSelector()
            sel.register(command_process.stdout, selectors.EVENT_READ)
            sel.register(command_process.stderr, selectors.EVENT_READ)
            while True:
                for key, _ in sel.select():
                    data = key.fileobj.read1().decode()
                    if not data:
                        return
                    if key.fileobj is command_process.stdout:
                        stdout_function(data)
                    else:
                        stderr_function(data)
        except BaseException as exc:
            exception_function(exc)
            command_process.terminate()
            raise

    control_process = multiprocessing.Process(target=target)
    control_process.start()
    return control_process



"""Utility code for working with Python's asyncio."""
import asyncio
import typing


def sync(
    awaitable: typing.Awaitable,
    event_loop: typing.Optional[asyncio.events.AbstractEventLoop] = None,
):
    """
    Run an awaitable synchronously. Good for calling asyncio code from sync Python.

    The usage is as follows:

    .. highlight:: python
    .. code-block:: python

        from ori.asyncio import sync

        async def your_function(arg1, arg2):
            # Additional asyncio code can go here
            return arg1 + arg2

        # If you call your_function() directly, you will get a coroutine
        # object. You need to either use the await keyword...
        # or the sync() wrapper as below.

        value = sync(your_function(1, 2))
        assert value == 3

    Args:
        awaitable: This is a function declared with `async def` or another type
            of awaitable object in Python.
        event_loop: An :mod:`asyncio` event loop. This defaults to `None`, in which
            case this function will try to pull the running event loop.
            If no event loop exists, this function will try to create one.

    Returns:
        This returns the same value that `awaitable` would if you
        used the `await` keyword instead of :func:`sync`.

    Raises:
        TypeError: If `awaitable` is not actually an awaitable object,
            then we raise a TypeError.
    """
    if not event_loop:
        event_loop = asyncio.get_event_loop()
    return event_loop.run_until_complete(awaitable)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ori']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ori',
    'version': '0.1.1',
    'description': 'Concurrency tools for Python.',
    'long_description': 'Ori, a high-level concurrency library for Python\n=================================================\n\nOri is a high-level wrapper around Python\'s `concurrent.futures` module, designed to make multithreading and multiprocessing as easy as possible.\n\n\nFrequently Asked Questions (FAQs)\n---------------------------------\n\n**Who made Ori?**\n\nOri was written by `James Mishra <https://jamesmishra.com>`_ and incubated at `Neocrym <https://neocrym.com>`_, a record label that uses artificial intelligence to find and promote musicians. Neocrym heavily relies on Ori to make their I/O-bound Python code run faster.\n\nThe source code for Ori is owned by Neocrym Records Inc, but licensed to Ori under the MIT License.\n\n**Why should I use Ori over directly interfacing with `concurrent.futures`?**\n\nThe Python module `concurrent.futures` was introduced as a high-level abstraction over lower-level interfaces like `threading.Thread` and `multiprocessing.Process`. However, `concurrent.futures` merely moves the problem away from managing threads or processes to managing *executors*. Ori has the ambitious goal of also abstracting away the executors--making multithreading or multiprocessing no harder than writing single-threaded code.\n\n**Is Ori a good replacement for Python\'s asyncio?**\n\nFor the hardcore `asyncio` user, probably not. Ori is focused on providing high-level abstractions over Python\'s `concurrent.futures` module that provides speed boosts for synchronous, I/O-bound Python.\n\n**Where did the name "Ori" come from?**\n\nThe name "Ori" is a reference to the god-like villains in the Stargate TV shows. There is no meaningful connection between the villains or concurrency.\n',
    'author': 'James Mishra',
    'author_email': 'james.mishra@neocrym.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
